@@ -23,6 +23,7 @@ void Rf_error(const char*, ...);
 #include "git2/diff.h"
 #include "git2/submodule.h"
 #include "git2/sys/index.h"
+#include "git2/sys/filter.h"
 
 #include "refs.h"
 #include "repository.h"
@@ -1379,22 +1380,109 @@ static int mkpath2file(
 	return error;
 }
 
-static int buffer_to_file(
+struct checkout_stream {
+	git_writestream base;
+	const char *path;
+	int fd;
+	int open;
+};
+
+static int checkout_stream_write(
+	git_writestream *s, const char *buffer, size_t len)
+{
+	struct checkout_stream *stream = (struct checkout_stream *)s;
+	int ret;
+
+	if ((ret = p_write(stream->fd, buffer, len)) < 0)
+		giterr_set(GITERR_OS, "Could not write to '%s'", stream->path);
+
+	return ret;
+}
+
+static int checkout_stream_close(git_writestream *s)
+{
+	struct checkout_stream *stream = (struct checkout_stream *)s;
+	assert(stream && stream->open);
+
+	stream->open = 0;
+	return p_close(stream->fd);
+}
+
+static void checkout_stream_free(git_writestream *s)
+{
+	GIT_UNUSED(s);
+}
+
+static int blob_content_to_file(
 	checkout_data *data,
 	struct stat *st,
-	git_buf *buf,
+	git_blob *blob,
 	const char *path,
-	mode_t file_mode)
+	const char *hint_path,
+	mode_t entry_filemode)
 {
-	int error;
+	int flags = data->opts.file_open_flags;
+	mode_t file_mode = data->opts.file_mode ?
+		data->opts.file_mode : entry_filemode;
+	git_filter_options filter_opts = GIT_FILTER_OPTIONS_INIT;
+	struct checkout_stream writer;
+	mode_t mode;
+	git_filter_list *fl = NULL;
+	int fd;
+	int error = 0;
+
+	if (hint_path == NULL)
+		hint_path = path;
 
 	if ((error = mkpath2file(data, path, data->opts.dir_mode)) < 0)
 		return error;
 
-	if ((error = git_futils_writebuffer(
-			buf, path, data->opts.file_open_flags, file_mode)) < 0)
+	if (flags <= 0)
+		flags = O_CREAT | O_TRUNC | O_WRONLY;
+	if (!(mode = file_mode))
+		mode = GIT_FILEMODE_BLOB;
+
+	if ((fd = p_open(path, flags, mode)) < 0) {
+		giterr_set(GITERR_OS, "Could not open '%s' for writing", path);
+		return fd;
+	}
+
+	filter_opts.attr_session = &data->attr_session;
+	filter_opts.temp_buf = &data->tmp;
+
+	if (!data->opts.disable_filters &&
+		(error = git_filter_list__load_ext(
+			&fl, data->repo, blob, hint_path,
+			GIT_FILTER_TO_WORKTREE, &filter_opts)))
 		return error;
 
+	/* setup the writer */
+	memset(&writer, 0, sizeof(struct checkout_stream));
+	writer.base.write = checkout_stream_write;
+	writer.base.close = checkout_stream_close;
+	writer.base.free = checkout_stream_free;
+	writer.path = path;
+	writer.fd = fd;
+	writer.open = 1;
+
+	error = git_filter_list_stream_blob(fl, blob, (git_writestream *)&writer);
+
+	assert(writer.open == 0);
+
+	git_filter_list_free(fl);
+
+	if (error < 0)
+		return error;
+
+	if (GIT_PERMS_IS_EXEC(mode)) {
+		data->perfdata.chmod_calls++;
+
+		if ((error = p_chmod(path, mode)) < 0) {
+			giterr_set(GITERR_OS, "Failed to set permissions on '%s'", path);
+			return error;
+		}
+	}
+
 	if (st) {
 		data->perfdata.stat_calls++;
 
@@ -1402,53 +1490,11 @@ static int buffer_to_file(
 			giterr_set(GITERR_OS, "Error statting '%s'", path);
 			return error;
 		}
-	}
 
-	if (GIT_PERMS_IS_EXEC(file_mode)) {
-		data->perfdata.chmod_calls++;
-
-		if ((error = p_chmod(path, file_mode)) < 0)
-			giterr_set(GITERR_OS, "Failed to set permissions on '%s'", path);
-	}
-
-	return error;
-}
-
-static int blob_content_to_file(
-	checkout_data *data,
-	struct stat *st,
-	git_blob *blob,
-	const char *path,
-	const char * hint_path,
-	mode_t entry_filemode)
-{
-	mode_t file_mode = data->opts.file_mode ?
-		data->opts.file_mode : entry_filemode;
-	git_buf out = GIT_BUF_INIT;
-	git_filter_list *fl = NULL;
-	int error = 0;
-
-	if (hint_path == NULL)
-		hint_path = path;
-
-	if (!data->opts.disable_filters)
-		error = git_filter_list__load_with_attr_session(
-			&fl, data->repo, &data->attr_session, blob, hint_path,
-			GIT_FILTER_TO_WORKTREE, GIT_FILTER_OPT_DEFAULT);
-
-	if (!error)
-		error = git_filter_list_apply_to_blob(&out, fl, blob);
-
-	git_filter_list_free(fl);
-
-	if (!error) {
-		error = buffer_to_file(data, st, &out, path, file_mode);
 		st->st_mode = entry_filemode;
-
-		git_buf_free(&out);
 	}
 
-	return error;
+	return 0;
 }
 
 static int blob_content_to_link(
@@ -1966,6 +2012,7 @@ static int checkout_write_merge(
 	git_merge_file_result result = {0};
 	git_filebuf output = GIT_FILEBUF_INIT;
 	git_filter_list *fl = NULL;
+	git_filter_options filter_opts = GIT_FILTER_OPTIONS_INIT;
 	int error = 0;
 
 	if (data->opts.checkout_strategy & GIT_CHECKOUT_CONFLICT_STYLE_DIFF3)
@@ -2015,9 +2062,12 @@ static int checkout_write_merge(
 		in_data.ptr = (char *)result.ptr;
 		in_data.size = result.len;
 
-		if ((error = git_filter_list__load_with_attr_session(
-				&fl, data->repo, &data->attr_session, NULL, git_buf_cstr(&path_workdir),
-				GIT_FILTER_TO_WORKTREE, GIT_FILTER_OPT_DEFAULT)) < 0 ||
+		filter_opts.attr_session = &data->attr_session;
+		filter_opts.temp_buf = &data->tmp;
+
+		if ((error = git_filter_list__load_ext(
+				&fl, data->repo, NULL, git_buf_cstr(&path_workdir),
+				GIT_FILTER_TO_WORKTREE, &filter_opts)) < 0 ||
 			(error = git_filter_list_apply_to_data(&out_data, fl, &in_data)) < 0)
 			goto done;
 	} else {
